"use strict";
/*
 * Copyright 2019 Google LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const bigquery = require("@google-cloud/bigquery");
const schema_1 = require("./schema");
const snapshot_1 = require("./snapshot");
const tracker_1 = require("../tracker");
const logs = require("../logs");
var schema_2 = require("./schema");
exports.RawChangelogSchema = schema_2.RawChangelogSchema;
exports.RawChangelogViewSchema = schema_2.RawChangelogViewSchema;
/**
 * An FirestoreEventHistoryTracker that exports data to BigQuery.
 *
 * When the first event is received, it creates necessary BigQuery resources:
 * - Dataset: {@link FirestoreBigQueryEventHistoryTrackerConfig#datasetId}.
 * - Table: Raw change log table {@link FirestoreBigQueryEventHistoryTracker#rawChangeLogTableName}.
 * - View: Latest view {@link FirestoreBigQueryEventHistoryTracker#rawLatestView}.
 * If any subsequent data export fails, it will attempt to reinitialize.
 */
class FirestoreBigQueryEventHistoryTracker {
    constructor(config) {
        this.config = config;
        this.initialized = false;
        this.bq = new bigquery.BigQuery();
    }
    record(events) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.initialize();
            const options = {
                raw: true,
            };
            const rows = events.map((event) => {
                return {
                    insertId: event.eventId,
                    json: {
                        timestamp: event.timestamp,
                        event_id: event.eventId,
                        document_name: event.documentName,
                        operation: tracker_1.ChangeType[event.operation],
                        data: JSON.stringify(event.data),
                    },
                };
            });
            yield this.insertData(rows, options);
        });
    }
    /**
     * Inserts rows of data into the BigQuery raw change log table.
     */
    insertData(rows, options) {
        return __awaiter(this, void 0, void 0, function* () {
            const payload = {
                skipInvalidRows: false,
                ignoreUnkownValues: false,
                rows: rows,
            };
            try {
                const dataset = this.bq.dataset(this.config.datasetId);
                const table = dataset.table(this.rawChangeLogTableName());
                logs.dataInserting(rows.length);
                yield table.insert(payload, options);
                logs.dataInserted(rows.length);
            }
            catch (e) {
                // Reinitializing in case the destintation table is modified.
                this.initialized = false;
                throw e;
            }
        });
    }
    /**
     * Creates the BigQuery resources with the expected schema for {@link FirestoreEventHistoryTracker}.
     * After the first invokation, it skips initialization assuming these resources are still there.
     */
    initialize() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.initialized) {
                return;
            }
            yield this.initializeDataset();
            yield this.initializeRawChangeLogTable();
            yield this.initializeLatestView();
            this.initialized = true;
        });
    }
    /**
     * Creates the specified dataset if it doesn't already exists.
     */
    initializeDataset() {
        return __awaiter(this, void 0, void 0, function* () {
            const dataset = this.bq.dataset(this.config.datasetId);
            const [datasetExists] = yield dataset.exists();
            if (datasetExists) {
                logs.bigQueryDatasetExists(this.config.datasetId);
            }
            else {
                logs.bigQueryDatasetCreating(this.config.datasetId);
                yield dataset.create();
                logs.bigQueryDatasetCreated(this.config.datasetId);
            }
            return dataset;
        });
    }
    /**
     * Creates the raw change log table if it doesn't already exist.
     * TODO: Validate that the BigQuery schema is correct if the table does exist,
     */
    initializeRawChangeLogTable() {
        return __awaiter(this, void 0, void 0, function* () {
            const changelogName = this.rawChangeLogTableName();
            const dataset = this.bq.dataset(this.config.datasetId);
            const table = dataset.table(changelogName);
            const [tableExists] = yield table.exists();
            if (tableExists) {
                logs.bigQueryTableAlreadyExists(table.id, dataset.id);
            }
            else {
                logs.bigQueryTableCreating(changelogName);
                const options = {
                    // `friendlyName` needs to be here to satisfy TypeScript
                    friendlyName: changelogName,
                    schema: schema_1.RawChangelogSchema,
                };
                yield table.create(options);
                logs.bigQueryTableCreated(changelogName);
            }
            return table;
        });
    }
    /**
     * Creates the latest snapshot view, which returns only latest operations
     * of all existing documents over the raw change log table.
     */
    initializeLatestView() {
        return __awaiter(this, void 0, void 0, function* () {
            const dataset = this.bq.dataset(this.config.datasetId);
            const view = dataset.table(this.rawLatestView());
            const [viewExists] = yield view.exists();
            if (viewExists) {
                logs.bigQueryViewAlreadyExists(view.id, dataset.id);
            }
            else {
                const latestSnapshot = snapshot_1.latestConsistentSnapshotView(this.config.datasetId, this.rawChangeLogTableName());
                logs.bigQueryViewCreating(this.rawLatestView(), latestSnapshot.query);
                const options = {
                    friendlyName: this.rawLatestView(),
                    view: latestSnapshot,
                };
                yield view.create(options);
                yield view.setMetadata({ schema: schema_1.RawChangelogViewSchema });
                logs.bigQueryViewCreated(this.rawLatestView());
            }
            return view;
        });
    }
    rawChangeLogTableName() {
        return `${this.config.tableId}_raw_changelog`;
    }
    rawLatestView() {
        return `${this.config.tableId}_raw_latest`;
    }
}
exports.FirestoreBigQueryEventHistoryTracker = FirestoreBigQueryEventHistoryTracker;
